@@ -1,6 +1,6 @@
 /* This is the machine-specific part for MIPS R[2346810]000 processors
 
-  Copyright (C) 1995,1996,1997,1998,2000,2003,2005,2007 Free Software Foundation, Inc.
+  Copyright (C) 1995,1996,1997,1998,2000,2003,2005,2007,2010 Free Software Foundation, Inc.
 
   This file is part of Gforth.
 
@@ -26,10 +26,16 @@
 #endif
 
 /* cache flush stuff */
-#ifdef ultrix
+#if defined(ultrix)
 #include <mips/cachectl.h>
+#elif (defined(__OpenBSD__) || defined(__NetBSD__) || defined(__FreeBSD__))
+# if (SIZEOF_VOID_P == 4)
+#  include <mips/sysarch.h>
+# else
+#  include <mips64/sysarch.h>
+# endif
 #else
-/* works on Irix */
+/* works on Irix and Android */
 #include <sys/cachectl.h>
 #endif
 
@@ -46,3 +52,7 @@
 #define CFAREG asm("$20")
 #define TOSREG asm("$21")
 #endif /* FORCE_REG */
+
+int check_prim(char *start, int len);
+
+#define CHECK_PRIM(start,len) check_prim(start,len)
