@@ -1,6 +1,6 @@
 /* Gforth support functions
 
-  Copyright (C) 1995,1996,1997,1998,2000,2003,2004,2006,2007,2008 Free Software Foundation, Inc.
+  Copyright (C) 1995,1996,1997,1998,2000,2003,2004,2006,2007,2008,2009,2010,2011,2012,2013,2014,2015 Free Software Foundation, Inc.
 
   This file is part of Gforth.
 
@@ -40,57 +40,57 @@
 #if defined(HAVE_LIBDL) || defined(HAVE_DLOPEN) /* what else? */
 #include <dlfcn.h>
 #endif
+#ifdef HAS_DEBUG
+extern int debug;
+# define debugp(x...) do { if (debug) fprintf(x); } while (0)
+#endif
+#ifdef HAVE_MPROBE
+#include <mcheck.h>
+#endif
 
 #ifdef HAS_FILE
-char *cstr(Char *from, UCell size, int clear)
+char *cstr(Char *from, UCell size)
 /* return a C-string corresponding to the Forth string ( FROM SIZE ).
-   the C-string lives until the next call of cstr with CLEAR being true */
+   the C-string lives until free */
 {
-  static struct cstr_buffer {
-    char *buffer;
-    size_t size;
-  } *buffers=NULL;
-  static int nbuffers=0;
-  static int used=0;
-  struct cstr_buffer *b;
-
-  if (buffers==NULL)
-    buffers=malloc(0);
-  if (clear)
-    used=0;
-  if (used>=nbuffers) {
-    buffers=realloc(buffers,sizeof(struct cstr_buffer)*(used+1));
-    buffers[used]=(struct cstr_buffer){malloc(0),0};
-    nbuffers=used+1;
-  }
-  b=&buffers[used];
-  if (size+1 > b->size) {
-    b->buffer = realloc(b->buffer,size+1);
-    b->size = size+1;
-  }
-  memcpy(b->buffer,from,size);
-  b->buffer[size]='\0';
-  used++;
-  return b->buffer;
+  char * string = malloc(size+1);
+  memcpy(string,from,size);
+  string[size]='\0';
+  return string;
 }
 
-char *tilde_cstr(Char *from, UCell size, int clear)
+Cell negate(Cell n)
+{
+  return -n;
+}
+
+char *tilde_cstr(Char *from, UCell size)
 /* like cstr(), but perform tilde expansion on the string */
 {
-  char *s1,*s2;
-  int s1_len, s2_len;
+  char *s1,*s2,*s3;
+  int s1_len, s2_len, allocs1=0;
   struct passwd *getpwnam (), *user_entry;
 
   if (size<1 || from[0]!='~')
-    return cstr(from, size, clear);
+    return cstr(from, size);
   if (size<2 || from[1]=='/') {
     s1 = (char *)getenv ("HOME");
-    if(s1 == NULL)
-#if defined(_WIN32) || defined (MSDOS)
+    if((s1 == NULL) || access(s1, W_OK))
+#if defined(_WIN32) || defined (MSDOS) || defined(__CYGWIN__)
+      {
+	s1 = NULL;
+	s2 = (char *)getenv("HOMEDRIVE");
+	s3 = (char *)getenv("HOMEPATH");
+	if((s2 != NULL) && (s3 != NULL)) {
+	  asprintf(&s1, "%s%s", s2, s3);
+	  allocs1=1;
+	}
+      }
+    if((s1 == NULL) || access(s1, W_OK))
       s1 = (char *)getenv ("TEMP");
-      if(s1 == NULL)
-         s1 = (char *)getenv ("TMP");
-         if(s1 == NULL)
+    if((s1 == NULL) || access(s1, W_OK))
+      s1 = (char *)getenv ("TMP");
+    if((s1 == NULL) || access(s1, W_OK))
 #endif
       s1 = "";
     s2 = (char *)from+1;
@@ -100,7 +100,7 @@ char *tilde_cstr(Char *from, UCell size, int clear)
     for (i=1; i<size && from[i]!='/'; i++)
       ;
     if (i==2 && from[1]=='+') /* deal with "~+", i.e., the wd */
-      return cstr(from+3, size<3?0:size-3,clear);
+      return cstr(from+3, size<3?0:size-3);
     {
       char user[i];
       memcpy(user,from+1,i-1);
@@ -108,7 +108,7 @@ char *tilde_cstr(Char *from, UCell size, int clear)
       user_entry=getpwnam(user);
     }
     if (user_entry==NULL)
-      return cstr(from, size, clear);
+      return cstr(from, size);
     s1 = user_entry->pw_dir;
     s2 = (char *)from+i;
     s2_len = size-i;
@@ -120,7 +120,8 @@ char *tilde_cstr(Char *from, UCell size, int clear)
     char path[s1_len+s2_len];
     memcpy(path,s1,s1_len);
     memcpy(path+s1_len,s2,s2_len);
-    return cstr((Char *)path,s1_len+s2_len,clear);
+    if(allocs1) free(s1);
+    return cstr((Char *)path,s1_len+s2_len);
   }
 }
 
@@ -128,9 +129,9 @@ Cell opencreate_file(char *s, Cell wfam, int flags, Cell *wiorp)
 {
   Cell fd;
   Cell wfileid;
-  fd = open(s, flags|ufileattr[wfam], 0666);
+  fd = open(s, flags|ufileattr[wfam & 0x0F], (wfam >> 4) ^ 0666);
   if (fd != -1) {
-    wfileid = (Cell)fdopen(fd, fileattr[wfam]);
+    wfileid = (Cell)fdopen(fd, fileattr[wfam & 0x0F]);
     *wiorp = IOR(wfileid == 0);
   } else {
     wfileid = 0;
@@ -153,6 +154,19 @@ DCell timeval2us(struct timeval *tvp)
 #endif
 }
 
+DCell timespec2ns(struct timespec *tvp)
+{
+#ifndef BUGGY_LONG_LONG
+  return (tvp->tv_sec*(DCell)1000000000LL)+tvp->tv_nsec;
+#else
+  DCell d2;
+  DCell d1=mmul(tvp->tv_sec,1000000000);
+  d2.lo = d1.lo+tvp->tv_nsec;
+  d2.hi = d1.hi + (d2.lo<d1.lo);
+  return d2;
+#endif
+}
+
 DCell double2ll(Float r)
 {
 #ifndef BUGGY_LONG_LONG
@@ -230,9 +244,9 @@ Cell capscompare(Char *c_addr1, UCell u1, Char *c_addr2, UCell u2)
 
 struct Longname *listlfind(Char *c_addr, UCell u, struct Longname *longname1)
 {
-  for (; longname1 != NULL; longname1 = (struct Longname *)(longname1->next))
+  for (; longname1 != NULL; longname1 = LONGNAME_NEXT(longname1))
     if ((UCell)LONGNAME_COUNT(longname1)==u &&
-	memcasecmp(c_addr, (Char *)(longname1->name), u)== 0 /* or inline? */)
+	memcasecmp(c_addr, LONGNAME_NAME(longname1), u)== 0 /* or inline? */)
       break;
   return longname1;
 }
@@ -245,7 +259,7 @@ struct Longname *hashlfind(Char *c_addr, UCell u, Cell *a_addr)
     longname1=(struct Longname *)(a_addr[1]);
     a_addr=(Cell *)(a_addr[0]);
     if ((UCell)LONGNAME_COUNT(longname1)==u &&
-	memcasecmp(c_addr, (Char *)(longname1->name), u)== 0 /* or inline? */) {
+	memcasecmp(c_addr, LONGNAME_NAME(longname1), u)== 0 /* or inline? */) {
       return longname1;
     }
   }
@@ -259,7 +273,7 @@ struct Longname *tablelfind(Char *c_addr, UCell u, Cell *a_addr)
     longname1=(struct Longname *)(a_addr[1]);
     a_addr=(Cell *)(a_addr[0]);
     if ((UCell)LONGNAME_COUNT(longname1)==u &&
-	memcmp(c_addr, longname1->name, u)== 0 /* or inline? */) {
+	memcmp(c_addr, LONGNAME_NAME(longname1), u)== 0 /* or inline? */) {
       return longname1;
     }
   }
@@ -286,6 +300,122 @@ UCell hashkey1(Char *c_addr, UCell u, UCell ubits)
   return ukey;
 }
 
+#define ROL(x, n) ((x << n) | (x >> (64-n)))
+
+#define MIXKEY2 \
+  a1=ROL((a+(b^hashx[2]))*hashx[0],37)+hashx[4]; \
+  b1=ROL(((b-ROL(a,13))^hashx[3])*hashx[1],23)+hashx[5]; \
+  a^=a1; b^=b1
+
+uint64_t hashx[6] =
+  { 0x87c37b91114253d5ULL, 0x4cf5ad432745937fULL,
+    0x6c5f6f6cbe627173ULL, 0x7164c30603661c2fULL,
+    0xce5009401b441347ULL, 0x454fa335a6e63ad3ULL };
+
+void hashkey2(Char* c_addr, UCell u, uint64_t upmask, hash128 *h)
+{
+  // upmask is 0 for case sensitive, and 0x2020202020202020ULL for case insensitive
+  uint64_t a=h->a, b=h->b;
+  size_t pagesize=0x1000; /* may be smaller, but may not be larger than real pagesize */
+  uint64_t mixin, a1, b1;
+  Char* endp = c_addr+(u&-sizeof(uint64_t)), *endp1=c_addr+u-sizeof(uint64_t);
+  // read all full words
+  for(; c_addr<endp; c_addr+=sizeof(uint64_t)) {
+    memcpy(&mixin, c_addr, sizeof(uint64_t));
+    // printf("+%lx\n", mixin);
+    mixin |= upmask & ~(mixin >> 2); // case insensitive trick
+    a ^= mixin;
+    MIXKEY2;
+  }
+  u &= sizeof(uint64_t)-1;
+  if(u) {
+    // read last word
+    int shift    = (sizeof(uint64_t)-u)*8;
+    int lastshift= 0;
+    if(((intptr_t)(endp1+sizeof(uint64_t)-1) & (pagesize-1)) >= (pagesize-sizeof(uint64_t))) {
+      // last access might cross page size
+      endp = endp1;
+      lastshift= shift;
+    }
+    memcpy(&mixin, endp, sizeof(uint64_t));
+    // strip off parts read after the string end
+    // and mix in length of remaining fragment
+#ifdef WORDS_BIGENDIAN
+    mixin <<= lastshift;
+    mixin >>= shift;
+#else
+    mixin >>= lastshift;
+    mixin <<= shift;
+#endif
+    mixin |= upmask & ~(mixin >> 2); // case insensitive trick
+    // printf("+%lx\n", mixin);
+    a ^= mixin;
+    b ^= (uint64_t)u; // distinguisher for last block size goes into b
+    MIXKEY2;
+  }
+  MIXKEY2; // finalizing round
+
+  h->a = a; h->b = b;
+}
+
+UCell hashkey2a(Char *s, UCell n)
+{
+  UCell seed=n;
+  size_t w=sizeof(UCell);
+#define rotl(x,r) ((x << r) | (x >> (8*w - r)));
+  int rot1 = w*4-1;
+  size_t pagesize=4096; /* a power-of-2 that's at most as large as a page
+                           if you change this, change test/other.fs */
+  UCell k0 = w==8 ? 0xb64d532aaaaaaad5 : 0xb653aad5;
+  /* the 64-bit k0 gives not-so-catastrophic avalanche results on a
+     case-insensitive version; the 32-bit version is based on that */
+  UCell upmask = w==8 ? 0x2020202020202020L : 0x20202020L;
+  
+  UCell h,h1;
+  if (n<=w) {
+    UCell erase=(w-n)*8; /* bits to erase */
+    if (n==0)
+      return seed;
+    if (((((UCell)(s+n-1))^((UCell)(s+w-1)))&(-pagesize)) != 0) {
+      /* cell access would cross page boundary, but byte access wouldn't */
+      /* so cell access to s might incur a SIGSEGV */
+      memcpy(&h,(char *)(s+n-w),w);
+#ifdef WORDS_BIGENDIAN
+      h = h & ((~(UCell)0) >> erase);
+#else
+      h = h>>erase;
+#endif
+    } else {
+      memcpy(&h,s,w);
+#ifdef WORDS_BIGENDIAN
+      h = h>>erase;
+#else
+      h = h & ((~(UCell)0) >> erase);
+#endif
+    }
+    h |= upmask & ~(h >> 2); // case insensitive trick
+    h = (h^seed)*k0;
+    h = rotl(h,rot1);
+    return h*k0;
+  } else {
+    Char *p = s;
+    h = seed;
+    do {
+      memcpy(&h1,p,w);
+      h1 |= upmask & ~(h1 >> 2); // case insensitive trick
+      h = (h^h1)*k0;
+      h = rotl(h,rot1);
+      p += w;
+    } while (p<s+n-w);
+    p = s+n-w;
+    memcpy(&h1,p,w);
+    h1 |= upmask & ~(h1 >> 2); // case insensitive trick
+    h = (h^h1)*k0;
+    h = rotl(h,rot1);
+    return h*k0;
+  }
+}
+
 struct Cellpair parse_white(Char *c_addr1, UCell u1)
 {
   /* use !isgraph instead of isspace? */
@@ -309,11 +439,14 @@ struct Cellpair parse_white(Char *c_addr1, UCell u1)
 #ifdef HAS_FILE
 Cell rename_file(Char *c_addr1, UCell u1, Char *c_addr2, UCell u2)
 {
-  char *s1=tilde_cstr(c_addr2, u2, 1);
-  return IOR(rename(tilde_cstr(c_addr1, u1, 0), s1)==-1);
+  char *s1=tilde_cstr(c_addr2, u2);
+  char *s2=tilde_cstr(c_addr1, u1);
+  return IOR(rename(s2, s1)==-1);
+  free(s1);
+  free(s2);
 }
 
-struct Cellquad read_line(Char *c_addr, UCell u1, Cell wfileid)
+struct Cellquad read_line(Char *c_addr, UCell u1, FILE *wfileid)
 {
   UCell u2, u3;
   Cell flag, wior;
@@ -322,13 +455,24 @@ struct Cellquad read_line(Char *c_addr, UCell u1, Cell wfileid)
 
   flag=-1;
   u3=0;
+  if (u1>0)
+    gf_regetc(wfileid);
   for(u2=0; u2<u1; u2++) {
-    c = getc((FILE *)wfileid);
+    do{
+      c = getc(wfileid);
+      if((wior=FILEIO(ferror(wfileid)))) clearerr(wfileid);
+    } while (wior == TOIOR(EINTR));
+    if(wior) break;
     u3++;
     if (c=='\n') break;
     if (c=='\r') {
-      if ((c = getc((FILE *)wfileid))!='\n')
-	ungetc(c,(FILE *)wfileid);
+      do {
+	c = getc(wfileid);
+	if((wior=FILEIO(ferror(wfileid)))) clearerr(wfileid);
+      } while(wior==TOIOR(EINTR));
+      if(wior) break;
+      if (c!='\n')
+	gf_ungetc(c,wfileid);
       else
 	u3++;
       break;
@@ -339,7 +483,6 @@ struct Cellquad read_line(Char *c_addr, UCell u1, Cell wfileid)
     }
     c_addr[u2] = (Char)c;
   }
-  wior=FILEIO(ferror((FILE *)wfileid));
   r.n1 = u2;
   r.n2 = flag;
   r.n3 = u3;
@@ -352,7 +495,7 @@ struct Cellpair file_status(Char *c_addr, UCell u)
   struct Cellpair r;
   Cell wfam;
   Cell wior;
-  char *filename=tilde_cstr(c_addr, u, 1);
+  char *filename=tilde_cstr(c_addr, u);
 
   if (access (filename, F_OK) != 0) {
     wfam=0;
@@ -378,10 +521,11 @@ struct Cellpair file_status(Char *c_addr, UCell u)
   }
   r.n1 = wfam;
   r.n2 = wior;
+  free(filename);
   return r;
 }
 
-Cell to_float(Char *c_addr, UCell u, Float *rp)
+Cell to_float(Char *c_addr, UCell u, Float *rp, Char dot)
 {
   /* convertible string := <significand>[<exponent>]
      <significand> := [<sign>]{<digits>[.<digits0>] | .<digits> }
@@ -418,9 +562,9 @@ Cell to_float(Char *c_addr, UCell u, Float *rp)
   aftersign: 
   if (s >= send)
     goto exponent;
-  switch (c=*s) {
+  if((c=*s)==dot) { *t++ = '.'; ndots++; s++; goto aftersign; }
+  switch (c) {
   case '0' ... '9': *t++ = c; ndigits++; s++; goto aftersign;
-  case '.':         *t++ = c; ndots++;   s++; goto aftersign;
   default:                                    goto exponent;
   }
  exponent:
@@ -500,14 +644,37 @@ UCell rshift(UCell u1, UCell n)
 }
 
 #ifndef STANDALONE
+#ifdef HAVE_MPROBE
+void gforth_abortmcheck(enum mcheck_status reason)
+{
+  throw(-2049-reason);
+}
+#endif
+
+void gforth_free(void * ptr)
+{
+#ifdef HAVE_MPROBE
+  if(debug_mcheck) {
+    int reason=mprobe(ptr);
+    debugp(stderr, "free(%8p)=%d;\n", ptr, reason);
+    if(reason > 0) {
+      throw(-2049-reason);
+    }
+  }
+#endif
+  free(ptr);
+}
+
 int gforth_system(Char *c_addr, UCell u)
 {
   int retval;
   char *prefix = getenv("GFORTHSYSTEMPREFIX") ? : DEFAULTSYSTEMPREFIX;
   size_t prefixlen = strlen(prefix);
   char buffer[prefixlen+u+1];
+  int MAYBE_UNUSED old_tp;
+  fflush(stdout);
 #ifndef MSDOS
-  int old_tp=terminal_prepped;
+  old_tp=terminal_prepped;
   deprep_terminal();
 #endif
   memcpy(buffer,prefix,prefixlen);
@@ -538,24 +705,39 @@ void gforth_ms(UCell u)
 
 UCell gforth_dlopen(Char *c_addr, UCell u)
 {
-  char * file=tilde_cstr(c_addr, u, 1);
+  char * file=tilde_cstr(c_addr, u);
   UCell lib;
 #if defined(HAVE_LIBLTDL)
   lib = (UCell)lt_dlopen(file);
+  free(file);
   if(lib) return lib;
 #elif defined(HAVE_LIBDL) || defined(HAVE_DLOPEN)
 #ifndef RTLD_GLOBAL
 #define RTLD_GLOBAL 0
 #endif
-  lib = (UCell)dlopen(file, RTLD_GLOBAL | RTLD_LAZY);
+  lib = (UCell)dlopen(file, RTLD_GLOBAL);
+  free(file);
   if(lib) return lib;
+  fprintf(stderr, "%s\n", dlerror());
 #elif defined(_WIN32)
   lib = (UCell) GetModuleHandle(file);
+  free(file);
   if(lib) return lib;
 #endif
   return 0;
 }
 
+void gforth_dlclose(UCell lib)
+{
+#if defined(HAVE_LIBLTDL)
+  (void)lt_dlclose((lt_dlhandle)lib);
+#elif defined(HAVE_LIBDL) || defined(HAVE_DLOPEN)
+  dlclose((void*)lib);
+#elif defined(_WIN32)
+  FreeLibrary(lib);
+#endif
+}
+
 #endif /* !defined(STANDALONE) */
 
 
